/*
  CPP_CONTEST=2014
  CPP_PROBLEM=B
  CPP_LANG=C+OPENMP+MPI
  CPP_PROCESSES_PER_NODE=marte 1
*/

/* RECORD
Albert Segura Salvador, Pedro Benedicte Illescas 
students of the Technical University of Catalunya
In the IV Spanish Parallel Programming Contest
time 6169 msec
speed-up 4.94
*/ 

/*Only OpenMP parallelism is used (as for problem A).
The work in the matrix is assigned dynamically to the threads,
and the threads accumulate their results in local matrices,
which are accumulated in parallel using half the total number of threads.
Higher speed-up could be obtained with message passing and using
the two nodes marte+mercurio.
*/

#include <stdlib.h>
#include <stdio.h>
#include <omp.h>
#define NUM_THREADS 6

void accumulatelocal(int *a,int N,int v,int r1,int c1,int r2,int c2)
{
  int i,j;
  int *pos;

  for(i=r1;i<=r2;i++)
  {
    pos=&a[i*N];
    for(j=c1;j<=c2;j++)
    {
      pos[j]+=v;
    }
  }
}

void copymatrix(int t,double *o,double *d)
{
  int i;

  for(i=0;i<t*t;i++)
  {
    d[i]=o[i];
  }
}

void obtainsubmatrix(int size,int row,int col,int t,double *o,double *d)
{
  int i,j;
  int temprow,tempcol;
  for(i=0;i<size;i++)
  {
    for(j=0;j<size;j++)
    {
       temprow=((i+row<t) ? (i+row):(i+row-t));
       tempcol=((j+col<t) ? (j+col):(j+col-t));
       d[i*size+j]=o[temprow*t+tempcol];
    }
  }
}

void matrixmultiplication(int t,double *sm,double *r)
{
  int i,j,k;
  double temp;

  for(i=0;i<t;i++)
  {
    for(j=0;j<t;j++)
    {
       temp=0;
       for(k=0;k<t;k++)
       {
          temp+=sm[i*t+k]*sm[k*t+j];
       }
       r[i*t+j]=temp;
    }
  }
}

void writeresult(int size,double *r,int row,int col,int t,double *m)
{
  int i,j;

  for(i=0;i<size;i++)
  {
    for(j=0;j<size;j++)
    {
       m[((i+row<t) ? (i+row):(i+row-t))*t+((j+col<t) ? (j+col):(j+col-t))]+=r[i*size+j];
    }
  }
}

void writematrix(int t,double *m)
{
  int i,j;

  for(i=0;i<t;i++)
  {
    for(j=0;j<t;j++)
    {
      printf("%.2f ",m[i*t+j]);
    }
    printf("\n");
  }
}

// a = a + b
void addMatrix(double * a, double * b, int m) {
  int i, j;
  for (i = 0; i < m; ++i) {
    for (j = 0; j < m; ++j) {
      a[i*m+j] = a[i*m+j] + b[i*m+j];
    }
  }

}

void  sec(int t,double *a,int *b,int nodo,int np)
{
  int i;


  double * c_array[NUM_THREADS];
  for (i=0;i<NUM_THREADS; i++) {
    c_array[i]=(double *) calloc(sizeof(double),t*t);
  }
	if (nodo == 0) {
  #pragma omp parallel for schedule(dynamic,5) num_threads(NUM_THREADS)
  for(i=0;i<t*t;i++) {
        int size;
        double *c; //matrix where to accumulate the squares of the submatrices
        double *sm; //to store the submatrix to multiply
        double *r; //to store the matrix result
        c = c_array[omp_get_thread_num()];
        size=b[i];
        sm=(double *) malloc(sizeof(double)*size*size);
        r=(double *) calloc(sizeof(double),size*size);
        obtainsubmatrix(size,i/t,i%t,t,a,sm);
  #ifdef DEBUG
        printf("Submatrix %d, %d: \n",i/t,i%t);
        writematrix(size,sm);
  #endif
        matrixmultiplication(size,sm,r);
  #ifdef DEBUG
        printf("Updated submatrix %d, %d: \n",i/t,i%t);
        writematrix(size,r);
  #endif
        writeresult(size,r,i/t,i%t,t,c);
  #ifdef DEBUG
        printf("Updated %d, %d: \n",i/t,i%t);
        writematrix(t,c);
  #endif
        free(sm);
        free(r);
  }
  for (i=0;i<NUM_THREADS;++i) {
    addMatrix(a,c_array[i],t);
    free(c_array[i]);
  }
  }
/*
  omp_set_num_threads(NUM_THREADS/2);
  #pragma omp task
  {
    addMatrix(c_array[omp_get_thread_num()*2],c_array[omp_get_thread_num()*2+1],t);
    if (omp_get_thread_num()%2 == 0) {
      addMatrix(c_array[omp_get_thread_num()*2],c_array[omp_get_thread_num()*2+2],t);
      if (omp_get_thread_num() == 0) {
        addMatrix(a,c_array[0],t);
      } else if (omp_get_thread_num() == 4) {
        addMatrix(c_array[4],c_array[8],t);
      }
    }
  }
    addMatrix(a,c_array[4],t);*/
  //copymatrix(t,c,a);
#ifdef DEBUG
        printf("Updated %d, %d: \n",i/t,i%t);
        writematrix(t,a);
#endif
  //free(c);
}
